/***************************************************************************
 *   Copyright (C) 2003-2005 by Tom Bigwood, Ron Dicho                     *
 *   veediot@ansiscene.com, icedevil@icednet.org                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qbrush.h>
#include <qcursor.h>
#include <qimage.h>
#include <qpixmap.h>

#include "tundracanvas.h"
#include "tundracharpreview.h"
#include "tundracolors.h"
#include "charset.h"

TundraCanvas::TundraCanvas(TundraModel* m, TundraCommandProcessor* cp, int w, QWidget* parent, const char* name) : QWidget(parent, name),  top_line(0), model(m), command_processor(cp), ctrl_down(false), alt_down(false), shift_down(false), insert_on(false), key_pressed(false), cursor_move(false),  move_block(false), copy_block(false), current_charset(5), back_buffer(0), update_tl(QPoint(0,0)), update_br(QPoint(0,0)), selection1(0, 0),  selection2(0, 0), mouse_down(false), wid(w), cursor_on(true), mouse_y(1), mouse_x(1) {

    model->setCharSet(current_charset);

    setCursor(ibeamCursor);
    setBackgroundMode(Qt::NoBackground);
    setPaletteBackgroundColor(QColor(0,0,0));

    setFocusPolicy(QWidget::StrongFocus);
    setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::MinimumExpanding));
    debug_break = false;
    back_buffer = new QPixmap(wid,400);

    aspectChange();

    // Turn on mouse tracking
    setMouseTracking(true);

    QTimer* timer = new QTimer();
    connect(timer, SIGNAL(timeout()), this, SLOT(changeCursor()));
    timer->start(500);

    QTimer* scroll_timer = new QTimer();
    connect(scroll_timer, SIGNAL(timeout()), this, SLOT(scrollMouseOutOfBounds()));
    scroll_timer->start(30);

    QObject::connect(model, SIGNAL(aspectChanged()), this, SLOT(aspectChange()));
}


void TundraCanvas::setMoving() {
    QPoint top_left;
    QPoint bottom_right;

    model->getSelection(top_left, bottom_right);

    cursor_x = top_left.x();
    cursor_y = top_left.y();

    move_block = true;

    saved_selection_tl = top_left;
    saved_selection_br = bottom_right;

    move_command = command_processor->newMoveCommand(top_left, bottom_right);

    model->cancelSelection();

    emit moving();
}

void TundraCanvas::setCopying() {
    QPoint top_left;
    QPoint bottom_right;

    model->getSelection(top_left, bottom_right);

    cursor_x = top_left.x();
    cursor_y = top_left.y();

    copy_block = true;

    saved_selection_tl = top_left;
    saved_selection_br = bottom_right;

    // Copy the current selection
    saved_width = bottom_right.x() - top_left.x();
    saved_height = bottom_right.y() - top_left.y();

    saved_doc = new TundraDoc*[saved_height];
    saved_doc[0] = new TundraDoc[saved_width*saved_height];

    for(int i = 1; i < saved_height; i++) {
        saved_doc[i] = saved_doc[0]+(saved_width*i);
    }

    for(int i = 0; i < saved_height; i++) {
        for(int j = 0; j < saved_width; j++) {
            saved_doc[i][j].character = model->getItem(top_left.x()+j, top_left.y()+i)->character;
            saved_doc[i][j].fg_color = model->getItem(top_left.x()+j, top_left.y()+i)->fg_color;
            saved_doc[i][j].bg_color = model->getItem(top_left.x()+j, top_left.y()+i)->bg_color;
        }
    }

    model->cancelSelection();

    emit copying();
}

void TundraCanvas::scrollMouseOutOfBounds() {
    if(!copy_block && !move_block) {
        QMouseEvent e(QEvent::MouseMove, QPoint(mouse_x, mouse_y), Qt::NoButton, Qt::NoButton);
        mouseMoveEvent(&e);
    }
}

void TundraCanvas::changeCursor() {
    if(cursor_on)
        cursor_on = false;
    else
        cursor_on = true;

    repaint();
}

void TundraCanvas::setWidth(int width) {
    wid = width;

    delete back_buffer;
    back_buffer = new QPixmap(wid, 400);

    aspectChange();
}

void TundraCanvas::paintEvent(QPaintEvent*) {

    bool temp_cursor_move = cursor_move;

    int x, y, w, h = 0;

    // Bounding box for the update
    if((update_tl == update_br) && !cursor_move) {
        // We're resizing
        if(back_buffer != 0)
            delete back_buffer;

        back_buffer = new QPixmap(wid,height());
        back_buffer->fill(Qt::black);

        y = top_line;
        h = (height()/pix_height) + top_line;

        // Check to see if we're resizing when we
        // are already at the bottom line
        if(h > NUM_ROWS) {
            h = NUM_ROWS;
            y = NUM_ROWS-height()/pix_height;
            setTopLine(y);
        }

        x = 0;
        w = 80;
    } else if(cursor_move) {
        // The cursor moved
        x = model->getLastPosition().x();
        y = model->getLastPosition().y();
        w = x+1;
        h = y+1;

        cursor_move = false;
    } else {
        // We're painting only the affected region
        y = update_tl.y();
        h = update_br.y();
        x = update_tl.x();
        w = update_br.x();
    }

    QPainter paint(back_buffer);

    for(int i = y; i < h; i++) {
        for(int j = x; j < w; j++) {

            if(model->getItem(j,i)->character != 0) {

                TundraDoc* loc = model->getItem(j,i);

                __int64 id = loc->character;
                id = id << 24;
                id = id | loc->bg_color;
                id = (id << 32) | loc->fg_color;

                if(character_map.contains(id)) {

                    QPoint top_left;
                    QPoint bottom_right;

                    model->getSelection(top_left, bottom_right);

                    if(i >= top_left.y() && i < bottom_right.y() && j >= top_left.x() && j < bottom_right.x()) {

                        if(inverse_character_map.contains(id)) {
                            paint.drawPixmap(j*pix_width, (i-top_line)*pix_height, inverse_character_map[id]);
                        } else {
                            QColor fg_color = model->getPalette(model->getItem(j, i)->fg_color);
                            QColor bg_color = model->getPalette(model->getItem(j, i)->bg_color);

                            // Inverse the colors
                            fg_color.setRgb(255 - fg_color.red(), 255 - fg_color.green(), 255 - fg_color.blue());
                            bg_color.setRgb(255 - bg_color.red(), 255 - bg_color.green(), 255 - bg_color.blue());

                            QPixmap background(pix_width, pix_height);
                            QImage bgimage = model->getCharacter(model->getItem(j,i)->character)->convertToImage().convertDepth(8);
                            bgimage.setColor(0, fg_color.rgb());
                            bgimage.setColor(1, bg_color.rgb());
                            QImage scaled = bgimage.smoothScale(pix_width, pix_height);

                            background.convertFromImage(scaled);

                            paint.drawPixmap(j*pix_width, (i-top_line)*pix_height, background);
                            inverse_character_map.insert(id, background);
                        }
                    } else
                        paint.drawPixmap(j*pix_width, (i-top_line)*pix_height, character_map[id]);
                } else {
                    QPixmap background(pix_width, pix_height);
                    QImage bgimage = model->getCharacter(model->getItem(j,i)->character)->convertToImage().convertDepth(8);
                    bgimage.setColor(0, model->getPalette(model->getItem(j, i)->fg_color).rgb());
                    bgimage.setColor(1, model->getPalette(model->getItem(j, i)->bg_color).rgb());
                    QImage scaled = bgimage.smoothScale(pix_width, pix_height);

                    background.convertFromImage(scaled);

                    character_map.insert(id, background);
                    paint.drawPixmap(j*pix_width, (i-top_line)*pix_height, background);
                }
            } else {
                QColor color;

                if(model->hasSelection()) {
                    // We have a selection

                    QPoint top_left;
                    QPoint bottom_right;

                    model->getSelection(top_left, bottom_right);

                    if(i >= top_left.y() && i < bottom_right.y() && j >= top_left.x() && j < bottom_right.x()) {
                        // Inverse the colors
                        paint.drawPixmap(j*pix_width, (i-top_line)*pix_height, white_block);
                    }
                } else {
                    // Erase the cursor
                    //paint.drawPixmap(j*pix_width, (i-top_line)*pix_height, black_block);
                    paint.drawLine(j*pix_width, (i-top_line)*pix_height+pix_height-1, j*pix_width+pix_width-1, (i-top_line)*pix_height+pix_height-1);
                }
            }
        }
    }

    if(copy_block || move_block) {
        // Move the block and selection on the screen

        // Calculate the selection width and height
        int selection_width = saved_selection_br.x() - saved_selection_tl.x();
        int selection_height = saved_selection_br.y() - saved_selection_tl.y();

        if(move_block) {
            // Paint over the old block if we're moving
            paint.fillRect(saved_selection_tl.x()*pix_width, (saved_selection_tl.y()-top_line)*pix_height, selection_width*pix_width, selection_height*pix_height, QBrush(Qt::black));
        }

        // Draw the block (inversed) in the new location
        for(int i = 0; i < selection_height ; i++) {
            for(int j = 0; j < selection_width; j++) {
                TundraDoc* current = NULL;

                if(move_block)
                    current = model->getItem(j+saved_selection_tl.x(),i+saved_selection_tl.y());
                else if(copy_block)
                    current = &saved_doc[i][j];

                __int64 id = current->character;
                id = id << 24;
                id = id | current->bg_color;
                id = (id << 32) | current->fg_color;

                if(current->character != 0)
                    paint.drawPixmap((j+cursor_x)*pix_width, ((i+cursor_y)-top_line)*pix_height, inverse_character_map[id]);
                else
                    paint.drawPixmap((j+cursor_x)*pix_width, ((i+cursor_y)-top_line)*pix_height, white_block);
            }
        }
    }

    QColor fg_color;
    fg_color = model->getPalette(model->getCursorItem()->fg_color);
    paint.setPen(QColor(255-fg_color.red(), 255-fg_color.green(), 255-fg_color.blue()));

    // Draw the blinking cursor
    TundraDoc* current = model->getItem(model->getCurrentPosition().x(), model->getCurrentPosition().y());

    __int64 id = current->character;
    id = id << 24;
    id = id | current->bg_color;
    id = (id << 32) | current->fg_color;

    if(cursor_on || temp_cursor_move) {
        paint.setPen(QColor(255, 255, 255));
    } else {
        paint.setPen(QColor(0, 0, 0));
    }

    paint.drawLine(model->getCurrentPosition().x()*pix_width, (model->getCurrentPosition().y()-top_line)*pix_height+pix_height-1, model->getCurrentPosition().x()*pix_width+pix_width-1, (model->getCurrentPosition().y()-top_line)*pix_height+pix_height-1);

    update_tl = update_br = QPoint(0,0);

    bitBlt(this, 0, 0, back_buffer, 0, 0, wid, height(),CopyROP,false);

}

void TundraCanvas::focusInEvent(QFocusEvent*) {
    alt_down = false;
    ctrl_down = false;
}


void TundraCanvas::keyPressEvent(QKeyEvent* e) {
    bool temp_copy_block = copy_block;

    int my_key = e->key();
    if((!copy_block && !move_block) && (my_key == Key_Enter || my_key == Key_Return)) {
        cursor_move = true;
        model->setCurrentY(model->getCurrentPosition().y()+1);

        if(model->getCurrentPosition().y() >= (top_line + height()/pix_height))
            setTopLine(top_line+1);
        else
            repaint(false);
    }

    if(e->state() & AltButton)
        alt_down = true;
    else
	alt_down = false;

    if(e->state() & ControlButton)
        ctrl_down = true;
    else 
        ctrl_down = false;

    if(my_key == Key_Insert) {
        insert_on = !insert_on;
    } else if((my_key >= 0x1030 && my_key < 0x103a) || my_key == Key_Space || (e->text().at(0).unicode() > 31 && e->text().at(0).unicode() < 128) || my_key == Key_Backspace || (my_key >= Key_1 && my_key <= Key_9)) {
        if(alt_down) {
            if((my_key >= 0x31 && my_key < 0x3A) || my_key == Key_0) {
				if(my_key == Key_0)
					current_charset = 9;
				else
					current_charset = my_key - 0x31;
                model->setCharSet(current_charset);
            }
            if(my_key == Key_U) {
                model->setCurrentFg(model->getCursorItem()->fg_color);
                model->setCurrentBg(model->getCursorItem()->bg_color);
            }
        } else if(ctrl_down) {
            if(my_key >= Key_1 && my_key <= Key_7) {
                current_charset = (my_key - 0x31) + 10;
                model->setCharSet(current_charset);
            }
        } else {
            QPoint loc = model->getCurrentPosition();
            unsigned char character = 0;
            int fg_color;
            int bg_color;

            model->setMaxY(model->getCurrentPosition().y());

            fg_color = model->getCurrentFg();
            bg_color = model->getCurrentBg();

            if(e->ascii() < 128 && e->ascii() > 31) {
                character = e->text().at(0).unicode();
            } else if(my_key >= 0x1030 && my_key < 0x103a) {
                character = char_set[current_charset][my_key - 0x1030];
            } else if(my_key == Key_Backspace) {
                character = 0;
                int x = loc.x();
                x--;
                if(x < 0)
                    x = 0;

                loc.setX(x);

                fg_color = 0;
                bg_color = 0;
            }

            key_pressed = true;

            TundraCommand* command = NULL;

            if(!model->hasSelection() && !copy_block && !move_block) {
                if(my_key == Key_Delete)
                    command = command_processor->newDeleteCommand(loc.x(), loc.y());
                else {
                    if(insert_on && my_key != Key_Backspace)
                        command = command_processor->newInsertCommand(loc.x(), loc.y(), character, bg_color, fg_color);
                    else {
                        TundraCharacterCommand* char_command = command_processor->newCharacterCommand(loc.x(),
                                                               loc.y(), character, bg_color, fg_color);
                        if(my_key == Key_Backspace)
                            char_command->setBackspace();

                        command = char_command;
                    }
                }
            } else if(copy_block) {
                // Stamp
                if(my_key == Key_S) {
                    command = paste();
                }
                if(my_key == Key_P || my_key == Key_Space) {
                    copy_block = false;
                    command = paste();
                }
            } else if(move_block) {
                if(my_key == Key_Space) {
                    QPoint start_position(cursor_x, cursor_y);
                    move_block = false;
                    move_command->setDestination(start_position);
                    command_processor->process(move_command);

                    emit movingDone();
                }
            } else {
                QPoint top_left;
                QPoint bottom_right;
                model->getSelection(top_left, bottom_right);

                if(ctrl_down) {
                    // Here we want to fill the current selection with the requested character
                    TundraFillBothStrategy* strategy = new TundraFillBothStrategy(character, fg_color, bg_color);
                    TundraBlockCommand* block_command = command_processor->newBlockCommand(top_left, bottom_right, strategy);

                    command = block_command;
                    model->cancelSelection();
                } else if(my_key == Key_E) {
                    TundraEraseStrategy* strategy = new TundraEraseStrategy();
                    TundraBlockCommand* block_command = command_processor->newBlockCommand(top_left, bottom_right, strategy);

                    command = block_command;
                    model->cancelSelection();
                } else if(my_key == Key_X) {
                    TundraFlipXStrategy* strategy = new TundraFlipXStrategy();
                    TundraBlockCommand* block_command = command_processor->newBlockCommand(top_left, bottom_right, strategy);

                    command = block_command;
                    model->cancelSelection();
                } else if(my_key == Key_Y) {
                    TundraFlipYStrategy* strategy = new TundraFlipYStrategy();
                    TundraBlockCommand* block_command = command_processor->newBlockCommand(top_left, bottom_right, strategy);

                    command = block_command;
                    model->cancelSelection();
                } else if(my_key == Key_F) {
                    emit fillRequested();
                } else if(my_key == Key_C) {
                    setCopying();
                } else if(my_key == Key_M) {
                    setMoving();
                }

            }

            if(command != NULL)
                command_processor->process(command);


            if(temp_copy_block && (my_key == Key_P || my_key == Key_Space)) {
                delete [] saved_doc[0];
                delete [] saved_doc;

                emit copyingDone();
            }

            key_pressed = false;

            // Fix this - no need to do 2 repaints
            if(model->getCurrentPosition().y() >= (top_line + height()/pix_height) || model->getCurrentPosition().y() < top_line)
                setTopLine(model->getCurrentPosition().y());
        }
    } else if((!model->hasSelection() || shift_down)) {
        switch(e->key()) {
        case Key_Home:
            model->setCurrentPosition(0, model->getCurrentPosition().y());
            if(shift_down) {
                int y = model->getCurrentPosition().y();
                if(y >= selection1.y())
                    y += 1;
                int x = model->getCurrentPosition().x();
                if(x >= selection1.x())
                    x += 1;
                changeSelection(x, y);
            }
            cursor_move = true;
            repaint(false);
            break;
        case Key_End:
            model->setCurrentPosition(79, model->getCurrentPosition().y());
            if(shift_down) {
                int y = model->getCurrentPosition().y();
                if(y >= selection1.y())
                    y += 1;
                int x = model->getCurrentPosition().x();
                if(x >= selection1.x())
                    x += 1;
                changeSelection(x, y);
            }
            cursor_move = true;
            repaint(false);
            break;
        case Key_Left:
            if(ctrl_down) {
                if(model->getCurrentBg() == 0) {
                    model->setCurrentBg((NUM_COLORS/2)-1);
                } else
                    model->setCurrentBg(model->getCurrentBg()-1);
            } else if(copy_block || move_block) {
                cursor_x -= 1;

                repaint(false);
            } else {
                cursor_move = true;
                if(model->getCurrentPosition().x()-1 >= 0)
                    model->setCurrentX(model->getCurrentPosition().x()-1);
                else
                    model->setCurrentX(0);
                
                if(shift_down) {
                    int y = model->getCurrentPosition().y();
                    if(y >= selection1.y())
                        y += 1;

                    int x = model->getCurrentPosition().x();
                    if(x >= selection1.x())
                        x += 1;

                    changeSelection(x, y);
                }
                repaint(false);
            }
            break;
        case Key_Right:
            if(ctrl_down) {
                model->setCurrentBg((model->getCurrentBg()+1)%(NUM_COLORS/2));
            } else if(copy_block || move_block) {
                cursor_x += 1;

                repaint(false);
            } else {
                cursor_move = true;
                model->setCurrentX(model->getCurrentPosition().x()+1);

                if(shift_down) {
                    int y = model->getCurrentPosition().y();
                    if(y >= selection1.y())
                        y += 1;
                    int x = model->getCurrentPosition().x();
                    if(x >= selection1.x())
                        x += 1;
                    changeSelection(x, y);
                }

                repaint(false);
            }
            break;
        case Key_Up:
            if(ctrl_down) {
                model->setCurrentFg((model->getCurrentFg()+1)%NUM_COLORS);
            } else if(copy_block || move_block) {
                cursor_y -= 1;

                if(cursor_y < top_line)
                    setTopLine(cursor_y);

                repaint(false);
            } else {
                cursor_move = true;
                if(model->getCurrentPosition().y()-1 >= 0)
                    model->setCurrentY(model->getCurrentPosition().y()-1);
                else
                    model->setCurrentY(0);

                if(shift_down) {
                    int y = model->getCurrentPosition().y();
                    if(y >= selection1.y())
                        y += 1;
                    int x = model->getCurrentPosition().x();
                    if(x >= selection1.x())
                        x += 1;
                    changeSelection(x, y);
                }

                if(model->getCurrentPosition().y() < top_line)
                    setTopLine(model->getCurrentPosition().y());
                else
                    repaint(false);
            }
            break;
        case Key_Down:
            if(ctrl_down) {
                if(model->getCurrentFg() == 0) {
                    model->setCurrentFg(NUM_COLORS-1);
                } else
                    model->setCurrentFg(model->getCurrentFg()-1);
            } else if(copy_block || move_block) {
                cursor_y += 1;

                if(cursor_y >= (top_line + height()/pix_height))
                    setTopLine(top_line+1);

                repaint(false);
            } else {
                cursor_move = true;
                model->setCurrentY(model->getCurrentPosition().y()+1);

                if(shift_down) {
                    int y = model->getCurrentPosition().y();
                    if(y >= selection1.y())
                        y += 1;
                    int x = model->getCurrentPosition().x();
                    if(x >= selection1.x())
                        x += 1;
                    changeSelection(x, y);
                }

                if(model->getCurrentPosition().y() >= (top_line + height()/pix_height))
                    setTopLine(top_line+1);
                else
                    repaint(false);
            }
            break;
		case Key_PageDown: {
			int offset = model->getCurrentPosition().y() - top_line;

            if(top_line + height()/pix_height > NUM_ROWS-height()/pix_height)
                setTopLine(NUM_ROWS-height()/pix_height);
            else
                setTopLine(top_line+height()/pix_height);

			model->setCurrentY(top_line+offset);
			
            break;
		}
		case Key_PageUp: {
			int offset = model->getCurrentPosition().y() - top_line;

            if(top_line - height()/pix_height < 0)
                setTopLine(0);
            else
                setTopLine(top_line-height()/pix_height);

			model->setCurrentY(top_line+offset);
            break;
		}
        default:
            e->ignore();
            break;
        }
    }

    if(e->key() == Key_Shift)
        shift_down = true;

    // Reset the selection
    if(!shift_down) {
        selection1 = QPoint(model->getCurrentPosition().x(), model->getCurrentPosition().y());
    }
}

void TundraCanvas::keyReleaseEvent(QKeyEvent* e) {
    switch(e->key()) {
    case Key_Control:
        ctrl_down = false;
        break;
    case Key_Alt:
        alt_down = false;
        break;
    case Key_Shift:
        shift_down = false;
        break;
    }
}
bool TundraCanvas::setTopLine(int line) {
    if(line > -1 && line <= NUM_ROWS-height()/pix_height) {
        top_line = line;
        // Make sure we repaint the whole screen
        update_tl = update_br = QPoint(0,0);
        cursor_move = false;
        emit topLineChanged();
        return true;
    } else
        return false;
}

void TundraCanvas::mouseMoveEvent(QMouseEvent* e) {
    if(mouse_down && !copy_block && !move_block) {

        mouse_y = e->y();
        mouse_x = e->x();

        changeSelection(e->x()/pix_width, (e->y()/pix_height)+top_line);

    } else if(copy_block || move_block) {
        cursor_x = (e->x())/pix_width;
        cursor_y = ((e->y())/pix_height)+top_line;

        repaint(false);
    }
}

void TundraCanvas::mousePressEvent(QMouseEvent* e) {
    model->setCurrentPosition((e->x())/pix_width, ((e->y())/pix_height)+top_line);

    mouse_down = true;
    mouse_x = e->x();
    mouse_y = e->y();

    selection1 = QPoint(e->x()/pix_width, (e->y()/pix_height)+top_line);
    selection2 = selection1;

    model->setSelection(selection1, selection2);
    model->setModified();

    // Paste
    if(copy_block) {
        TundraCommand* command = paste();
        copy_block = false;

        command_processor->process(command);
        delete [] saved_doc[0];
        delete [] saved_doc;

        emit copyingDone();
    }
    if(move_block) {
        QPoint start_position(cursor_x, cursor_y);
        move_block = false;
        move_command->setDestination(start_position);
        command_processor->process(move_command);

        emit movingDone();
    }
}

void TundraCanvas::mouseReleaseEvent(QMouseEvent*) {
    mouse_down = false;
    mouse_y = 1;
    mouse_x = 1;
}

TundraBlockCommand* TundraCanvas::paste() {
    QPoint start_position(cursor_x, cursor_y);

    int saved_height = saved_selection_br.y() - saved_selection_tl.y();
    int saved_width = saved_selection_br.x() - saved_selection_tl.x();

    QPoint end_position(cursor_x+saved_width, cursor_y+saved_height);

    TundraPasteStrategy* strategy = new TundraPasteStrategy(saved_doc, saved_width, saved_height);
    TundraBlockCommand* block_command = command_processor->newBlockCommand(start_position, end_position, strategy);

    return block_command;
}

void TundraCanvas::aspectChange() {
    pix_width = wid/80;
    pix_height = int((wid/model->getAspect()) + .5);

    white_block = QPixmap(pix_width, pix_height);
    black_block = QPixmap(pix_width, pix_height);

    QPainter paint1(&white_block);
    QPainter paint2(&black_block);
    paint1.fillRect(0, 0, pix_width, pix_height, QBrush(Qt::white));
    paint2.fillRect(0, 0, pix_width, pix_height, QBrush(Qt::black));

    character_map.clear();
    inverse_character_map.clear();

    model->setModified();
}

void TundraCanvas::repaint(int x, int y, int w, int h, bool erase) {
    erase = false;
    QWidget::repaint(x, y, w, h, false);
}

void TundraCanvas::repaint() {
    QWidget::repaint(false);
}

void TundraCanvas::repaint(bool erase) {
    erase = false;
    QWidget::repaint(false);
}

void TundraCanvas::repaint(const QRect& r, bool erase) {
    erase = false;
    QWidget::repaint(r, false);
}

void TundraCanvas::repaint(const QRegion& reg, bool erase) {
    erase = false;
    QWidget::repaint(reg, false);
}

void TundraCanvas::focusOutEvent(QFocusEvent* e) {
    if(e->lostFocus())
        alt_down = false;
}

void TundraCanvas::wheelEvent(QWheelEvent* e) {
    if(e->delta() < 0) {
        setTopLine(top_line+qApp->wheelScrollLines());
    } else if(e->delta() > 0) {
        if(top_line < qApp->wheelScrollLines())
            setTopLine(0);
        else
            setTopLine(top_line-qApp->wheelScrollLines());
    }
}

void TundraCanvas::changeSelection(int x, int y) {
    selection2 = QPoint(x, y);
    QPoint top_left;
    QPoint bottom_right;

    // Find the top most point
    if(selection1.y() <= selection2.y()) {
        top_left.setY(selection1.y());
        bottom_right.setY(selection2.y());
    } else {
        top_left.setY(selection2.y());
        bottom_right.setY(selection1.y()+1);
    }

    // Find the left most point
    if(selection1.x() <= selection2.x()) {
        top_left.setX(selection1.x());
        bottom_right.setX(selection2.x());
    } else {
        top_left.setX(selection2.x());
        bottom_right.setX(selection1.x()+1);
    }

    // Scroll down
    if(y > (getTopLine()+(height()/pix_height)))
        setTopLine(top_line+1);

    // Scroll up
    if(y <= getTopLine())
        setTopLine(top_line-1);

    model->setSelection(top_left, bottom_right);
    model->setModified();
}
