/***************************************************************************
 *   Copyright (C) 2003-2005 by Tom Bigwood, Ron Dicho                     *
 *   veediot@ansiscene.com, icedevil@icednet.org                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef TUNDRACANVAS_H
#define TUNDRACANVAS_H

#include <qmap.h>
#include <qpainter.h>
#include <qpoint.h>
#include <qtimer.h>
#include <qwidget.h>
#include <qevent.h>
#include <qapplication.h>

#include "tundramodel.h"
#include "tundracommandprocessor.h"

#ifndef _WINDOWS
typedef long long __int64; 	 
#endif

//! The canvas onto which a drawing is rendered
/*!
    This custom widget handles displaying the characters that represent a drawing. All of
    the rendering is done manually, by creating a QPixmap to represent a character/color 
    combination, and then calling bitBlt to render that to the widget's paint device. 
    It handles key presses and mouse events to enable drawing to the canvas. It is contained 
    within a TundraCanvasGroup, which allows for scrolling of the canvas. This canvas is 
    reusable, and can be resized to any width, although the width should remain a multiple 
    of the number of columns in the current mode (80 for ANSi.)
 
    \sa TundraCanvasGroup
    \sa TundraModel
*/
class TundraCanvas : public QWidget {

    Q_OBJECT

public:
    //! Constructor
    /*!
        Constructor for Tundra Canvas
        \param m A pointer to the model
        \param cp A pointer to the main command processor
    */
    TundraCanvas(TundraModel* m, TundraCommandProcessor* cp, int w, QWidget* parent=0, const char* name=0);

    //! Standard Qt paintEvent
    /*!
     We override the widget's paint event so that we can do our own custom drawing. This is where all
     of the drawing for the canvas is done.
    */
    void paintEvent (QPaintEvent*);

    //! Standard Qt keyPressEvent
    /*!
     We override the widget's key press event so that we can capture all of the key presses related to drawing, such
     as the F-keys and (m)ove, (c)opy, (e)rase, etc.
    */
    void keyPressEvent(QKeyEvent* e);

    //! Standard Qt keyReleaseEvent
    /*!
     We override the widget's key release event so that we know when alt and ctrl are released
    */
    void keyReleaseEvent(QKeyEvent* e);

    //! Standard Qt wheelEvent
    /*!
     We override wheelEvent so that we can enable wheel mouse scrolling for the canvas
    */
    void wheelEvent(QWheelEvent* e);

    //! Get the canvas's current top line
    int getTopLine() const {
        return top_line;
    }

    //! Set the canvas's top line
    /*!
     This moves the canvas's view so that the line number passed in at
     the top
     \param line The line number to use as the top line
     \sa getTopLine
    */
    bool setTopLine(int line);

    //! Standard Qt mousePressEvent
    /*!
     We override mousePressEvent so that we know when to start a selection
    */
    void mousePressEvent(QMouseEvent* e);

    //! Standard Qt mouseMoveEvent
    /*!
     We override mouseMoveEvent so that we can size a selection or move a selected
     block when moving or copying
    */
    void mouseMoveEvent(QMouseEvent* e);

    //! Standard Qt mouseReleaseEvent
    /*!
     We override mouseReleaseEvent so that we know when to stop sizing a selection
    */
    void mouseReleaseEvent(QMouseEvent*);

    //! Check if a key is currently pressed
    bool isKeyPress() const {
        return key_pressed;
    }

    //! Get the bounding box for the current region to update
    /*!
     \param tl A QPoint in which to store the top left point
     \param br A QPoint in which to store the bottom right point
     \sa setBounds
    */
    void getBounds(QPoint& tl, QPoint &br) {
        tl = update_tl;
        br = update_br;
    }

    //! Set the bounding box for the current region to update
    /*!
     \param tl The QPoint to use for the top left point of the bounding box
     \param br The QPoint to use for the bottom right point of the bounding box
     \sa getBounds
    */
    void setBounds(QPoint tl, QPoint br) {
        update_tl = tl;
        update_br = br;
    }

    //! Standard Qt sizeHint
    /*!
     This is used by Qt to know how to size the widget to its preferred size
    */
    QSize sizeHint() const {
        return QSize(wid,400);
    }

    //! Create a paste command based on the current saved selection
    TundraBlockCommand* paste();

    //! Set the width of the canvas
    /*!
     \param width The new width of the canvas
    */
    void setWidth(int width);

    //! Standard Qt focusOutEvent
    /*!
     We override the focusOutEvent to force any alt/ctrl keypress to be voided (used to fix alt tab bug)
    */
    void focusOutEvent(QFocusEvent* e);

    //! Standard Qt focusInEvent
    /*!
     We override the focusInEvent to force any alt/ctrl keypress to be voided (used to fix alt tab bug)
    */
    void focusInEvent(QFocusEvent*);

    //! Creates a new TundraMoveCommand
    void setMoving();

    //! Saves the current selection for later pasting
    /*!
     \sa paste
    */
    void setCopying();

    //! Update the selection based on mouse or key moves
    /*!
        \param x The new x location on the canvas
        \param y The new y location on the canvas
    */
    void changeSelection(int x, int y);

public slots:
    //! Changes the height of each character by the aspect ratio in the model
    void aspectChange();

    //! Standard Qt repaint override
    /*!
     We override the repaint functions so we can force Qt not to ever erase the background
     of the widget
    */
    void repaint(int x, int y, int w, int h, bool erase = false);

    //! Standard Qt repaint override
    /*!
     We override the repaint functions so we can force Qt not to ever erase the background
     of the widget
    */
    void repaint();

    //! Standard Qt repaint override
    /*!
     We override the repaint functions so we can force Qt not to ever erase the background
     of the widget
    */
    void repaint(bool erase);

    //! Standard Qt repaint override
    /*!
     We override the repaint functions so we can force Qt not to ever erase the background
     of the widget
    */
    void repaint(const QRect& r, bool erase = false);

    //! Standard Qt repaint override
    /*!
     We override the repaint functions so we can force Qt not to ever erase the background
     of the widget
    */
    void repaint(const QRegion& reg, bool erase = false);

    //! Toggle whether the cursor should be drawn as white or black
    void changeCursor();

    //! Automatically scroll the canvas
    /*!
     We scroll the canvas based on a timer when the mouse is outside of the canvas and
     we have a selection
    */
    void scrollMouseOutOfBounds();
signals:
    //! Notify that the top line of the canvas has been changed
    void topLineChanged();

    //! Notify that the user is currently moving a block
    void moving();

    //! Notify that the user is currently copying a block
    void copying();

    //! Notify that the moving is finished
    void movingDone();

    //! Notify that the copying is finished
    void copyingDone();

    //! Notify that the user pressed the F key with a selection
    void fillRequested();
private:
    int top_line;
    TundraModel* model;
    TundraCommandProcessor* command_processor;
    bool debug_break;
    bool ctrl_down;
    bool alt_down;
    bool shift_down;
    bool insert_on;
    bool key_pressed;
    bool cursor_move;
    bool move_block;
    bool copy_block;
    unsigned char current_charset;
    QPixmap* back_buffer;
    QPoint update_tl;
    QPoint update_br;
    QMap<__int64,QPixmap> character_map;
    QMap<__int64,QPixmap> inverse_character_map;
    QPoint selection1;
    QPoint selection2;
    QPoint saved_selection_tl;
    QPoint saved_selection_br;
    int saved_width;
    int saved_height;
    QPixmap white_block;
    QPixmap black_block;
    bool mouse_down;
    int wid;
    int pix_height;
    int pix_width;
    int cursor_x;
    int cursor_y;
    TundraMoveCommand* move_command;
    TundraDoc** saved_doc;
    bool cursor_on;
    int mouse_y;
    int mouse_x;
};

#endif
