/***************************************************************************
 *   Copyright (C) 2003-2005 by Tom Bigwood, Ron Dicho                     *
 *   veediot@ansiscene.com, icedevil@icednet.org                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qabstractlayout.h>
#include <qclipboard.h>
#include <qfiledialog.h>
#include <qframe.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qhbox.h>
#include <qmainwindow.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qpoint.h>
#include <qscrollbar.h>
#include <qsplashscreen.h>
#include <qstatusbar.h>
#include <qstring.h>
#include <qtimer.h>
#include <qtoolbar.h>

#include "tundraaboutdialog.h"
#include "tundracanvas.h"
#include "tundracanvasgroup.h"
#include "tundracharpreview.h"
#include "tundrachatwindow.h"
#include "tundraclientsocket.h"
#include "tundracolors.h"
#include "tundraconnectdlg.h"
#include "tundraflipxstrategy.h"
#include "tundraflipystrategy.h"
#include "tundramodel.h"
#include "tundracurrentcolorbox.h"
#include "tundramimesource.h"
#include "tundrapalette.h"
#include "tundrapreview.h"
#include "tundrasave24bitstrategy.h"
#include "tundrasaveansistrategy.h"
#include "tundrasaveasciistrategy.h"
#include "tundraserversocket.h"
#include "tundrastartserverdlg.h"
#include "tundracommandprocessor.h"
#include "tundranetcommandvisitor.h"
#include "tundrauserlist.h"
#include "tundrafillchardlg.h"

extern unsigned char char_set[6][10];

//! The main window
/*!
	The TundraDrawWindow subclasses QMainWindow, which as you may expect, is the application's
	main window. It contains the canvas, all toolbars, menus, and a status bar.
*/
class TundraDrawWindow : public QMainWindow
{
    Q_OBJECT
public:
	//! Constructor
	/*!
		\param filename A filename passed in from a command line argument
		\param s A pointer to the splash screen so that we can update our progress and then close it
	*/
    TundraDrawWindow(QString filename, QSplashScreen* s, QWidget* parent = 0, const char* name = 0, WFlags f = WType_TopLevel );
    
	//! Standard Qt resizeEvent override
	/*!
		We override resizeEvent so that we can resize the canvas group
		accordingly
	*/
	void resizeEvent(QResizeEvent* e);
    
	//! Set the caption in the titlebar
	void setCaption(QString e);

public slots:
	//! The color mode has changed, update menu items, and possibly send the new mode to other clients
    void colorModeChanged() {
        if(model->getColorMode() == COLOR_24BIT)
            colorMode24Bit();
        else if(model->getColorMode() == COLOR_4BIT)
            colorMode4Bit();
    }

	//! Preview window was hidden, update menus
    void vgaPreviewHidden();

	//! Foreground color was changed, update the current color box
    void fgColorChanged();

	//! Background color was changed, update the current color box
    void bgColorChanged();

	//! A new document is going to come over the network
    void newDocumentSignaled();

	//! A new document has been received from the network
    void newDocumentFromNetwork();

	//! Create a new document
    void newDocument();

	//! Save the document
    void saveDocument();

	//! Save the document with a new file name
    void saveDocumentAs();

	//! Open a document
    void openDocument();

	//! The cursor was moved, update the coordinates in the status bar
    void cursorMoved();

	//! Show the about dialog
    void about();

	//! Fill with the current foreground and background colors
    void fillAttribute();

	//! Show the fill character dialog, then fill with the selected character
    void fillCharacter();

	//! Show the fill character dialog, then fill 
    void fillCharacterAttribute();

	//! Fill with the current foreground color
    void fillFgColor();

	//! Fill with the current background color
    void fillBgColor();

	//! Erase the current selection
    void erase();

	//! Start a quick copy
    void copy();

	//! Start a move
    void move();

	//! Flip horizontally the current selection
    void flipX();

	//! Flip vertically the current selection
    void flipY();

	//! Display the Fill popup menu on the canvas
    void showFillMenu();

	//! Copy the current selection to the clipboard
    void copyToClipboard();

	//! Cut the current selection to the clipboard
    void cutToClipboard();

	//! Paste from the clipboard at the current cursor position
    void pasteFromClipboard();

	//! Toggle the preview window
    void viewPreviewWindow();

	//! Show the start server dialog, start server
    void startServer();

	//! Show the connect to server dialog, connect to specified server
    void connectToServer();

	//! Handle any errors in attempting to connect to a server
    void clientSocketError(int e);

	//! Switch the color mode to 4-bit
    void colorMode4Bit();

	//! Switch the color mode to 24-bit
    void colorMode24Bit();

	//! Override the closeEvent so we can prompt for save if the document has changed
    void closeEvent(QCloseEvent* e);

	//! Disconnect from the server
    void disconnectFromServer();

	//! Shutdown the server
    void closeServer();

	//! Swap the foreground and background colors
    void swapFgBg();

	//! We have a selection, update the status bar and menus
    void selectionOn();

	//! The selection was cancelled, update the status bar and menus
    void selectionOff();

	//! A block is currently being moved, update the status bar with help
    void movingOn();

	//! The move was finished, update the status bar
    void movingOff();

	//! A block is being quick copied, update the status bar with help
    void copyingOn();

	//! The quick copy was finished, update the status bar
    void copyingOff();

	//! The contents of the clipboard have been modified, update the paste menu item
    void clipboardChanged();

	//! Undo the last command
    void undo();

	//! Redo the last undone command
    void redo();

	//! We are able to undo, update the menu item
    void canUndo();

	//! We are no longer able to undo, update the menu item
    void cantUndo();

	//! We are able to redo, update the menu item
	void canRedo() { edit_menu->setItemEnabled(redo_item_id, TRUE); }

	//! We are no longer able to undo, update the menu item
    void cantRedo() { edit_menu->setItemEnabled(redo_item_id, FALSE); }

	//! Toggle the dos aspect ratio emulation
    void emulateDosAspect() {
        if(model->getAspect() == 40.0) {
            model->setAspect(33.68);
            view_menu->setItemChecked(emulate_dos_id, true);

        }
        else {
            model->setAspect(40.0);
            view_menu->setItemChecked(emulate_dos_id, false);
        }
    }

	//! Hide the splash screen
    void hideSplash();

	//! Export the current state of the model as a data file for the unit tests
	void exportUnitTestData();

    //! Import a 24-bit unit test data file
    void importUnitTestData24();

    //! Import a 4-bit unit test data file
    void importUnitTestData4();

private:
    TundraPreview* preview;
    TundraModel* model;
    bool alt_f4_down;
    TundraCurrentColorBox* fgcolor_box;
    TundraCurrentColorBox* bgcolor_box;
    QStatusBar* status_bar;
    QLabel* messages;
    QLabel* line_number;
    QToolBar* palette_bar;
    QToolBar* picker_bar;
    QToolBar* charpreview_bar;
    QGridLayout* layout;
    QMenuBar* main_menu;
    QPopupMenu* file_menu;
    QPopupMenu* edit_menu;
    QPopupMenu* view_menu;
    QPopupMenu* tools_menu;
    QPopupMenu* help_menu;
    QPopupMenu* block_menu;
    QPopupMenu* fill_menu;
    QPopupMenu* mode_menu;
    QPopupMenu* key_fill_menu;
    QString caption;
    QString orig_caption;
    TundraServerSocket* socket_server;
    TundraClientSocket* client_socket;
    TundraCommandProcessor* command_processor;
    TundraNetCommandVisitor* net_command_visitor;
    TundraCanvasGroup* canvas_group;
    TundraChatWindow* chat_window;
    TundraUserList* user_list;
    QToolBar* chat_bar;
    QToolBar* user_bar;
    int emulate_dos_id;
    int preview_item_id;
    int undo_item_id;
    int redo_item_id;
    int save_item;
    int mode_4bit_id;
    int mode_24bit_id;
    int start_server_id;
    int connect_to_server_id;
    int disconnect_from_server_id;
    int close_server_id;
    int block_id;
    int move_id;
    int copy_id;
    int copy_clipboard_id;
    int cut_clipboard_id;
    int paste_clipboard_id;
    QSplashScreen* splash;
    QClipboard* clip_board;
};

