/***************************************************************************
 *   Copyright (C) 2003-2005 by Tom Bigwood, Ron Dicho                     *
 *   veediot@ansiscene.com, icedevil@icednet.org                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef TUNDRAMODEL_H
#define TUNDRAMODEL_H

#include <qcolor.h>
#include <qfile.h>
#include <qobject.h>
#include <qpixmap.h>
#include <qpoint.h>
#include <qmap.h>
#include <qmessagebox.h>
#include <qtextstream.h>

#include "tundracolors.h"
#include "tundradoc.h"
#include "tundrasavestrategy.h"
#include "tundrasave24bitstrategy.h"
#include "tundrasaveansistrategy.h"
#include "tundrasaveasciistrategy.h"
#include "utils.h"

#define NUM_CHARACTERS 256
#define COLOR_4BIT 0
#define COLOR_24BIT 1

//! The applications current document
/*!
    The model is used to represent the document in memory. It contains a
    2D array of TundraDocs that make up the canvas, along with several 
    methods for altering the document and getting information from it. It also
    contains other important application information, such as the current character
    set being used for the F-keys, cursor location, the current colors, and the 
    current aspect ratio.
*/
class TundraModel : public QObject {

    Q_OBJECT

public:
    //! Constructor
    TundraModel();

    //! Test if two documents are identical
    friend bool operator==(const TundraModel &a, const TundraModel &b);

    //! Get the actual color from a palette index
    /*!
        \param index The index (in 4-bit, this is 0-15, in 24-bit it is the packed integer).

        \return An actual QColor using the color from the index into the palette
    */
    QColor getPalette(int index) const;

    //! Get a QPixmap rendering of a character from the character set
    /*!
        \param index The ascii value of the character

        \return A pointer to a QPixmap rendering of the character
    */
    QPixmap* getCharacter(int index);

    //! Get a block from the canvas
    /*!
        \param x The x location of the block
        \param y The y location of the block

        \return A pointer to the TundraDoc representing the specified block
    */
    TundraDoc* getItem(int x, int y) const {
        if(x > -1 && x < NUM_COLS && y > -1 && y < NUM_ROWS)
            return &document[y][x];
        else
            return NULL;
    }

    //! Returns a pointer to the 2D array representing the canvas contents
    /*!
        \return A pointer to the 2D array representing the document
    */
	TundraDoc** getDocument() const { return document; }

    //! Change the current character set (F1-F10) to the specified index (1-6)
    /*!
        \param charset The index of the character set to use
    */
    void setCharSet(unsigned char charset) { current_charset = charset; emit charSetChanged(); }

    //! Get the index of the character set currently in use
    /*!
        \return The index of the character set in use
    */
    unsigned char getCharSet() const { return current_charset; }

    //! Set the current foreground color
    /*!
        \param color The index of the color to use as the current foreground color
    */
    bool setCurrentFg(int color);

    //! Set the current background color
    /*!
        \param color The index of the color to use as the current background color
    */
    bool setCurrentBg(int color);

    //! Set the current cursor position
    /*!
        \param x The x position of the cursor
        \param y The y location of the cursor

        \return TRUE if the cursor position is valid, FALSE otherwise
    */
    bool setCurrentPosition(int x, int y);

    //! Set the X position of the cursor
    /*!
        \param x The x position of the cursor

        \return TRUE if the cursor position is valid, FALSE otherwise
    */
    bool setCurrentX(int x);

    //! Set the Y position of the cursor
    /*!
        \param y The y position of the cursor

        \return TRUE if the cursor position is valid, FALSE otherwise
    */
    bool setCurrentY(int y);


    //! Get the character map used to flip characters for Flip X
    /*!
        \return A pointer to the character map
    */
    QMap<unsigned char, unsigned char>* getFlipXMap() { return &flip_x_map; }

    //! Get the character map used to flip character for Flip Y
    /*!
        \return A pointer to the character map
    */
    QMap<unsigned char, unsigned char>* getFlipYMap() { return &flip_y_map; }

    //! Set the color mode (optionally of a block that we are pasting)
    /*!
        This function switches the color mode, and converts the colors of blocks currently
        on the canvas to its representation in thew new color mode. 

        \param mode COLOR_4BIT or COLOR_24BIT
        \param doc A pointer to a block of characters (for use with paste)
        \param width The width of the block of characters (for use with paste)
        \param height The height of the block of characters (for use with paste)
    */
    void setColorMode(unsigned char mode, TundraDoc** doc = 0, int width = NUM_COLS, int height = NUM_ROWS);

    //! Get the color mode
    /*!
        \return COLOR_4BIT or COLOR_24BIT
    */
    unsigned char getColorMode() const { return color_mode; }

    //! WARNING: This function is obsolete, it should not be used.
    void setMaxY(int y) { if(y > max_y) max_y = y; }
    
    //! Get the current foreground color
    /*!
        \return The index of the current foreground color
    */
	int getCurrentFg() const { return current_fg; }

    //! Get the current background color
    /*!
        \return The index of the current background color
    */
    int getCurrentBg() const { return current_bg; }

    //! Get the current cursor position
    /*!
        \return A QPoint representing the cursor position
    */
    QPoint getCurrentPosition() const { return current_position; }

    //! Get the previous current position
    /*!
        \return A QPoint representing the previous cursor position
    */
    QPoint getLastPosition() const { return last_position; }

    //! Get the maximum Y position of the canvas
    int getMaxX() const { return max_x; }

    //! Get the maximum X position of the canvas
    int getMaxY() const { return max_y; }

    //! Get the TundraDoc representing the block under the cursor
    /*!
        \return A pointer to the TundraDoc representing the block under the cursor
    */
    TundraDoc* getCursorItem() { return &document[current_position.y()][current_position.x()]; }

    //! Emit a signal notifying that the document has been modified
    void setModified() { emit modified(); }

    //! Open a file
    /*!
        \param filename The name of the file to open
    */
    bool openDocument(QString filename);

    //! Open a 4-bit ansi
    /*!
        \param input A reference to the QFile that has been opened
        \param ba A pointer to a byte array in which to write the data in the case of copying to the clipboard
        \param doc A pointer to the block to copy to the clipboard
        \param width The width of the block to copy to the clipboard
        \param height The height of the block to copy to the clipboard
    */
    bool openANSi(QFile& input, QByteArray* ba = 0, TundraDoc** doc = 0, int width = NUM_COLS, int height = NUM_ROWS);

    //! Open a 4-bit ansi
    /*!
        \param input A reference to the QFile that has been opened
        \param ba A pointer to a byte array in which to write the data in the case of copying to the clipboard
        \param doc A pointer to the block to copy to the clipboard
        \param width The width of the block to copy to the clipboard
        \param height The height of the block to copy to the clipboard
    */
    bool open24BitANSi(QFile& input, QByteArray* ba = 0, TundraDoc** doc = 0, int width = NUM_COLS, int height = NUM_ROWS);

    //! Reset the document
    void newDocument();

    //! Save the document using the specified filename and save strategy
    /*!
        \param filename The filename to save as
        \param strategy A pointer to the save strategy to use
    */
    bool saveDocument(QString filename, TundraSaveStrategy* strategy);

    //! Save the document with the current filename and save strategy
    bool saveDocument();

    //! Emit a signal warning that a new document is going to come over the network
    void sendNewDocumentWarning() { emit warnNewDocument(); }

    //! Emit a signal notifying that we received a new document over the network
    void sendNewDocumentNotify() { emit newNetworkDocument(); }

    //! Return if we have a filename for saving
    /*!
        \return TRUE if we already have a filename, FALSE otherwise
    */
    bool hasFilename() const { return has_filename; }

    //! Get whether or not there is an active selection
    bool hasSelection() const { return ((selection_tl.x() != selection_br.x()) && (selection_tl.y() != selection_br.y())); }
    
    //! Get the active selection
    /*!
        \param top_left A reference to a QPoint in which to store the top left point of the selection
        \param bottom_right A reference to a QPoint in which to store the bottom right point of the selection
    */
    void getSelection(QPoint& top_left, QPoint& bottom_right) const {
        top_left = selection_tl;
        bottom_right = selection_br;
    }

    //! Change the aspect ratio
    /*!
        \param a The new aspect ratio (a to 1)
    */
    void setAspect(double a) {
        aspect = a;
        emit aspectChanged();
    }

    //! Get the current aspect ratio
    double getAspect() const {
        return aspect;
    }

    //! Set the active selection
    /*!
        \param top_left A QPoint representing the top left corner of the selection
        \param bottom_right A QPoint representing the bottom right corner of the selection
    */
    void setSelection(QPoint& top_left, QPoint& bottom_right) {
        bool had_selection = false;
        if(selection_tl != selection_br)
            had_selection = true;

        selection_tl = top_left;
        selection_br = bottom_right;

        // Bounds checking the selection
        if(selection_br.x() > (max_x+1))
            selection_br.setX(max_x+1);
        if(selection_br.y() > (max_y+1))
            selection_br.setY(max_y+1);
        if(selection_tl.x() < 0)
            selection_br.setX(0);
        if(selection_tl.y() < 0)
            selection_br.setX(0);

        if((selection_tl.x() != selection_br.x()) && (selection_tl.y() != selection_br.y())) {
            emit selectionOn();
        }
        else if(had_selection == true) {
            emit selectionOff();
        }
    }

    //! Turn off the active selection
    void cancelSelection() {
        selection_tl = selection_br = QPoint(0, 0);

        emit selectionOff();
    }

    //! Export the current document in a raw format for unit tests
    /*!
        \param filename The name of the raw data file
    */
    bool exportUnitTestData(QString filename);

    //! Import a raw unit test data file
    /*!
        \param filename The name of the raw data file
        \param color_mode The color mode of the data file we're opening (you should always know this)
    */
    bool importUnitTestData(QString filename, unsigned int color_mode);
signals:
    //! Notify that the aspect ratio has changed
    void aspectChanged();

    //! Notify that the model has been modified
    void modified();

    //! Notify that the foreground color has changed
    void fgColorChanged();

    //! Notify that the background color has changed
    void bgColorChanged();

    //! Notify that the cursor position has changed
    void positionChanged();

    //! Notify that the current character set has changed
    void charSetChanged();

    //! Notify that the color mode has changed
    void colorModeChanged();

    //! Warn that a new document is going to come over the network
    void warnNewDocument();

    //! Signal that a new document has been loaded from the network
    void newNetworkDocument();

    //! Notify that there is an active selection
    void selectionOn();

    //! Notify that there is no longer an active selection
    void selectionOff();
private:
    bool has_filename;
    QString save_filename;
    QPoint current_position;
    QPoint last_position;
    QPoint selection_tl;
    QPoint selection_br;
    int current_fg;
    int current_bg;
    int max_y;
    int max_x;
    unsigned char color_mode;
    unsigned char last_color_mode;
    unsigned char current_charset;
    QColor palette[NUM_COLORS];
    QPixmap characters[NUM_CHARACTERS];
    TundraDoc** document;
    double aspect;
    TundraSaveStrategy* current_save_strategy;
    QMap<unsigned char, unsigned char> flip_x_map;
    QMap<unsigned char, unsigned char> flip_y_map;
};


#endif
